/*
    Copyright (C) 2005, 2006  Vansteenhuyse Hubert <hvansteenhuyse@yahoo.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*******************************************************************************
    LOGIC LAYER
    -----------
        General application classes.

                                                            Sunday, 4 dec 2005.
*******************************************************************************/
#ifndef __CHexEd_H__
#define __CHexEd_H__
#include "internal.h"
#ifdef __WINDOWS__
    #include <windows.h>
#endif

const char BOX_HORZ = '-';
const char BOX_VERT = '|';
const char BOX_CORN = '+';


typedef enum DISPLAY_MODUS{
    DISP_NORMAL,
    DISP_REVERSE,
    DISP_HIGHLIGHTED
};

typedef enum ACTION{
    ACT_CANCEL = 0,
    ACT_PARSECOMMAND
};

typedef enum COMMANDID{
    CMD_UNKNOWN = 0,
    CMD_QUIT,
    CMD_HELP,
    CMD_LS,
    CMD_SET_WRITEMODE,
    CMD_SET_READMODE,
    CMD_GOTO
};

typedef enum SCREENID{
    SCR_NONE = 0,
    SCR_COMMAND,
    SCR_INTRO,
    SCR_BUFFER,
    SCR_CMD_LS,       //:ls command
    SCR_CMD_SEARCH,   // / command
    SCR_CMD_HELP
};

typedef enum KEYID{
    KNUL = 0,
    KCANCEL = 1000,
    KPGU,
    KPGD,
    KUP,
    KDOWN,
    KLEFT,
    KRIGHT,
    KHOME,
    KBACK
};

typedef enum EDITMODE{
    EDT_VIEW = 0,
    EDT_EDIT,
    EDT_COMMAND         //all ':xxx' commands
};

typedef struct DIRFILE{
    char FileName[1024];
    int  IsFolder;
}DIRFILE;
typedef struct ioDIR{   //implementation in main.cpp
    private:
        unsigned long   mElements;
        DIRFILE         **mContent;
        unsigned long   mSelectedItem, mTopItem;

        void QuickSort(unsigned long from, unsigned long to, int at = 0);


    public:
        CStr mSearchFor;
        char Directory[1024];
        KEYID mLastKey;
        ioDIR();
        ~ioDIR();

        void Reset();
        void AddElement(DIRFILE elem);
        DIRFILE* ElementAt(unsigned long at);
        unsigned long CountElements();

        void AddSelectedItem(unsigned long amount);
        void SubSelectedItem(unsigned long amount);
        void SetSelectedItem(unsigned long item);
        unsigned long GetSelectedItem();

        void AddTopItem(unsigned long amount);
        void SubTopItem(unsigned long amount);
        void SetTopItem(unsigned long item);
        unsigned long GetTopItem();

        unsigned long GetFirstFile(); //exclusion of . and .. folders

        void Sort(unsigned long from = 0, unsigned long to = 0, int at = 0);
}ioDIR;

class CHexEd;
class CObject{
    private:
        int mErrno;
    protected:
        void seterr(int e){ mErrno = e; }
    public:
        int err(){ return mErrno; }
        void noerr(){ mErrno = 0; }
};

class ioLowLevel : public CObject{
    private:
    protected:
        int ioIsEofFile(void* FileHandle);
        void* ioOpenFile(char *path, int swRO);
        void ioCloseFile(void* FileHandle);
        long ioReadFile(void* FileHandle, char *buffer, long amount);
        long ioWriteFile(void* FileHandle, char *buffer, long amount);
        void ioFlushFile(void* FileHandle);
        int ioSeekFile(void* FileHandle, unsigned long FileAddress);  //-1 = failed
        unsigned long ioFileSize(void* FileHandle);

    public:
        ioLowLevel();
        ~ioLowLevel();

        int GetKey();
        void DoEvents();

        //Fills struct dir with files like in specified path, errno is returned
        static int ioListDirectory(ioDIR *dir, char *path, char *like);
        static char ioGetFileSeparatorChar();
};

class CUserInput : public ioLowLevel{
    private:
    protected:
    public:
        CUserInput();
        ~CUserInput();

        int GetInput();
};

class CBuffer : public ioLowLevel{
    private:
      int             IsCopy;
      CStr            *mPath;
      char            *mBuf;
      int             mSwRO, mSwEof;
      long            mBufTail,
                      mBufSize;
      unsigned long   mFileAddress,
                      mFileSize;
      void            *mf;

    protected:

    public:
        CBuffer(char *path, int swRO = 1);  //swRO = 1 then file os opened as read only
        CBuffer(CBuffer *base);
        ~CBuffer();


        int IsEof(){ return ioIsEofFile(mf); }
        int IsReadOnly(){ return mSwRO; }
        void SetReadOnly(int sw){ mSwRO = sw; }
        char* GetPath(){ return (strlen(mPath->get()) == 0  ? (char*)"Empty Buffer" : mPath->get());}
        char* GetBuffer(){ return mBuf; }
        unsigned long GetFileSize(){ return mFileSize; }
        long GetBufferSize(){ return mBufSize; }
        long GetAmountLastLoaded(){ return mBufTail; }
        char* GetBufferTail(){ return mBuf + mBufTail; }
        unsigned long GetFileAddress(){ return mFileAddress; }
        long ReadFrom(unsigned long fileaddress, long amount);
        long WriteFile(unsigned long fileaddress, char *buf, long amount);
};

class uiLowLevel : public CObject{
    private:
        void *mWnd;
        DISPLAY_MODUS mModus;

    protected:
    public:
        uiLowLevel();
        ~uiLowLevel();

        void setDisplayModus(DISPLAY_MODUS modus);
        DISPLAY_MODUS getDisplayModus(){ return mModus; }

        void InitScreen();  //initialization
        void commit();      //commit changes to the screen

        void clrscr();
        void clrln();


        long GetMaxX();      //surface x (in chars)
        long GetMaxY();      //surface y (in chars)

        void cursor_on(int sw_turn);
        void gotoxy(int x, int y);     //move cursor

        void printchar(const char c);
        void print(const char *str);
        void printxy(int x, int y, const char *str);
        void printcharxy(int x, int y, const char c);

        char GetDisplayChar(char ch);   //converts a char to a '.' if it cannot be displayed on the screen
};

class CScreen : public uiLowLevel{
    private:

    protected:
    public:
      CScreen();
      ~CScreen();

      void CScreen::ShowError(unsigned int ErrorNumber, const char *fmt, ...);    //Shows an error message
      void clrprintxy(int x, int y, const char *fmt, ...);   //clears line smoothly, safe formatted output
      void right(int y, const char *fmt, ...);      //safe formatted output right
      void center(int y, const char *fmt, ...);      //safe formatted output center
      void linexy(int x, int y, int len);

      void DrawUnknownScreen(SCREENID id);
      void DrawIntro();
      void DrawLS(ioDIR *dir);
      void DrawSearch(CStr *search, int sw_hexside);
      void DrawCommandLine(char *cmd, int sw_reverse = 0);      //draws the command line

      long GetMaxHexDump();              //x location of end hexdump part on screen
      long GetMinAsciiDump();            //x location of begin ascii part on screen
      long GetHexDumpLineSize(); //computes the amount of chars required to fill a whole hexdump line
      long GetHexDumpScreenSize(); //computes the amount of chars required to fill a whole hexdump screen
      int HexDump(char *from, char *to, int swClear = 1);    //returns amount of lines dumped on the screen
      void DrawHelp();
};

class CHexEd : public CObject{
    private:
      CStr *min, *mSearch;
      CBuffer *mBuffer;
      CScreen *mScreen;
      CUserInput *mInput;
      SCREENID mScreenID;
      EDITMODE mEditMode;
      int SW_QUIT, mSW_hexside;
      long mx, my;
      unsigned long mTop ,  mCursor ;
      char  input[3];
      ioDIR *mDir;

        void save();
        ACTION doInput(CStr *appendto);
        void dolistfiles();
        int dosearchinput();    //return 1 if search is to be runned
        void dohelp();

        long GetBufferIndexXY(int x = -1, int y = -1);    //computes the index inside mBuffer buffer where my + mx are pointing to
        void SetCursor(int x, int y);
        void MoveTop(KEYID ky);

    protected:
        void DrawFooter();                            //draws the footer(%, address...)
        void HighLight(int sw_hi, int x, int y);

    public:
        CHexEd();
        ~CHexEd();

        int search(int sw_fwd);    //search from cursor: sw_fwd = 1 then forward else backwards
        int LoadFile(char *path);
        void SetQuit(){ SW_QUIT = 1; }
        int main(int argc, char *argv[]);
        void DrawScreen(SCREENID id = SCR_NONE);
        void SetFocus(int swGotFocus);
        COMMANDID parseCommand(char *str);
};

/********************************************************************
  Globals
        implemented in main.cpp
********************************************************************/
extern CHexEd hxe;
const float VERSION = 0.1;
extern char *intro[];
extern char *help[];
extern int win_kbd;
extern int SW_SCREENREADY, SW_GOTFOCUS, SW_DISKREAD;
#endif /*__CHexEd_H__*/
