/*
    Copyright (C) 2005, 2006  Vansteenhuyse Hubert <hvansteenhuyse@yahoo.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "CHexEd.h"

/***************************************************************************************************
** GLOBALS...
***************************************************************************************************/
char *intro[] = {
  " HEXADECIMAL FILE TOOL ",
  "$BUILD",
  "$VERSION",
  "",
  "HXE is distributed under the General Public Licence (Version 2)",
  "http://www.gnu.org/copyleft/gpl.html",
  "",
  "type  :q<enter>   to quit      ",
  "type  :h<enter>   for help     ",
  "type  :ls<enter>  to list files",
  0
};


char *help[] = {
" hxe help...",
" This tool allows you to modify large binary files.",
"",
" COLON COMMANDS:",
"  Commands are typed in the vi (gnu/linux editor) style.",
"  From the main screen (=welcome and hexdump)",
"   :ls..............lists files/folders",
"   :[address].......Jumps to address or %% of file (eg ':15' or ':50%%')",
"   :q or ZZ.........quits",
"   :w...............Enables writing mode of the buffer",
"   :r...............Disables writing mode of the buffer",
"   i or o...........overwrite mode (warning: changes are applied directly)",
"                   type the hex value in hex part of the screendump",
"                                       OR",
"                   the charachter in the text part of the screendump",
"   /................Specify search pattern",
"   n................search next occurence",
"   p................search previous occurence",
"",
"  From the filelist screen:",
"   :cd.............change directory (eg: ':cd c:' changes to drive c:).",
"",
" The escape key is used to cancel all activities.",
0
};

//create an instance of CHexEd (global variable, declared in CHecEd.h)
CHexEd hxe;

//windows keyboard char
int win_kbd = 0;
int SW_SCREENREADY = 0,   //implemented for windows:
                          //when creating the window MainWndProc calls (in uiLowLevel.cpp)
                          //hxe.DrawScreen() to soon. SW_SCREENREADY is set to 1
                          //in CHexEd.main it is then ok to call hxe.DrawScreen()...

    SW_GOTFOCUS = 1,      //window focus flag
    SW_DISKREAD = 0;      //reading from disk or not
/***************************************************************************************************
** implentation of structure ioDIR...
***************************************************************************************************/
ioDIR::ioDIR(){
    memset(this, 0, sizeof(ioDIR));
}

ioDIR::~ioDIR(){
    Reset();
}

void ioDIR::Reset(){
    unsigned long iter = 0;
    DIRFILE *df;
    if(mContent){
        while(iter < mElements){
            df = mContent[iter];
            if(df)free(df);
            mContent[iter] = 0;
            iter++;
        }
        free(mContent);
    }
    mContent = 0;
    mElements = 0;
    memset(Directory, 0, 1024);
    mSelectedItem = 1;
    mTopItem = 1;
    mLastKey = KNUL;
    mSearchFor.clear();
}

void ioDIR::AddElement(DIRFILE elem){
    DIRFILE *df;
    mElements++;
    if(!mContent)
        mContent = (DIRFILE**)malloc(sizeof(DIRFILE*) * mElements);
    else
        mContent = (DIRFILE**)realloc(mContent, sizeof(DIRFILE*) * mElements);

    if(mContent){
        mContent[mElements - 1] = (DIRFILE*)malloc(sizeof(DIRFILE));
        df = mContent[mElements - 1];
        sprintf(df->FileName, elem.FileName);
        df->IsFolder = elem.IsFolder;
    }
}

DIRFILE* ioDIR::ElementAt(unsigned long at){
    DIRFILE *df = 0;
    if(mContent && at < mElements) df = mContent[at];
    return df;
}

unsigned long ioDIR::CountElements(){ return mElements; }

void ioDIR::AddSelectedItem(unsigned long amount){
    if(mSelectedItem + amount >= mElements - 1)
        mSelectedItem = mElements - 1;
    else
        mSelectedItem += amount;
}

void ioDIR::SubSelectedItem(unsigned long amount){
    if(mSelectedItem - amount >= mElements ||  mSelectedItem - amount == 0)
        mSelectedItem = 1;
    else
        mSelectedItem -= amount;
}

void ioDIR::SetSelectedItem(unsigned long item){
    if(item && item < mElements) mSelectedItem = item;
}

unsigned long ioDIR::GetSelectedItem(){ return mSelectedItem; }

void ioDIR::AddTopItem(unsigned long amount){
    if(mTopItem + amount > mElements - 1)
        mTopItem = mElements - 1;
    else
        mTopItem += amount;
}

void ioDIR::SubTopItem(unsigned long amount){
    if(mTopItem - amount >= mElements || mTopItem - amount == 0)
        mTopItem = 1;
    else
        mTopItem -= amount;
}
void ioDIR::SetTopItem(unsigned long item){
    if(item && item < mElements) mTopItem = item;
}
unsigned long ioDIR::GetTopItem(){ return mTopItem; }

unsigned long ioDIR::GetFirstFile(){ //exclusion of . and .. folders
    unsigned long Top = 0;
    while(Top < mElements
            &&
          (
            (mContent[Top]->FileName[0] == '.' && mContent[Top]->FileName[0] == 0)
            ||
            (mContent[Top]->FileName[0] == '.' && mContent[Top]->FileName[0] == '.')
          )
    )Top++;
    return Top;
}

//QUICKSORT
void ioDIR::QuickSort(unsigned long from, unsigned long to, int at){
    unsigned char pivot = 0;
    unsigned long _from, _to, part;
    DIRFILE *swap;
        if(from >= to) return;
        if( to >= mElements) to--;
        _from = from;
        _to = to;
        part = (_from + _to) / 2;

        if(at <= (int)strlen(mContent[part]->FileName))
            pivot = LCASE(mContent[part]->FileName[at]);
        else
            pivot = 0;
        do{

            while(_from < mElements && (at <= (int)strlen(mContent[_from]->FileName) ? LCASE(mContent[_from]->FileName[at]) : 0) < pivot) _from++;
            while(_to < mElements && (at <= (int)strlen(mContent[_to]->FileName) ? LCASE(mContent[_to]->FileName[at]) : 0) > pivot) _to--;
            if(_to > mElements){
                _from = mElements;
                _to = mElements;
            }else{
                if(_from <= _to){
                    swap = mContent[_from];
                    mContent[_from] = mContent[_to];
                    mContent[_to] = swap;
                    _from++;
                    _to--;
                }
            }
        }while(_from <= _to);
        if(from < _to)QuickSort(from, _to);
        if(_from < to)QuickSort(_from, to);
}


void ioDIR::Sort(unsigned long from, unsigned long to, int at){
  //TE DOEN: sorteren op volledige naam
    if(at == 0){
        from = GetFirstFile();
        to = CountElements();
        QuickSort(from, to, at);
        from++;
    }
}

/***************************************************************************************************
** PROGRAM ENTRY...
***************************************************************************************************/
int main(int argc, char *argv[]){
    return hxe.main(argc, argv);
}
