/*
    Copyright (C) 2005, 2006  Vansteenhuyse Hubert <hvansteenhuyse@yahoo.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*******************************************************************************
    +++ WINDOWS +++
    PRESENTATION LAYER
    ------------------
        Specialized screen output routines, in occurence windows windowing system
        and Graphical Device Interface.

                                                            Friday, 6 jan 2006.
*******************************************************************************/
#include "../../logic/CHexEd.h"
#include <string.h>
#define CURSOR_HEIGHT 4

const unsigned long COL_BACK = RGB(0, 0, 0);
const unsigned long COL_FORE = RGB(0, 144, 0);
const unsigned long COL_HIGH = RGB(254, 254, 254);

LRESULT CALLBACK
MainWndProc(HWND hwnd, UINT nMsg, WPARAM wParam, LPARAM lParam)
{
    switch (nMsg) {
        case WM_SETFOCUS:
            hxe.SetFocus(1);
            return 0;
            break;

        case WM_KILLFOCUS:
            hxe.SetFocus(0);
            return 0;
            break;

        case WM_DESTROY:
        	hxe.SetQuit();
        	PostQuitMessage(0);
        	return 0;
        	break;

        case WM_PAINT:
        	//update the screen
        	if(SW_SCREENREADY) hxe.DrawScreen();
    	   break;

        case WM_CHAR:
        	win_kbd = (int)wParam;
        	return 0;
        	break;

        case WM_KEYDOWN:
        	switch (wParam) {
        	case VK_PRIOR:
        	    win_kbd = (int)KPGU;
        	    break;
        	case VK_NEXT:
        	    win_kbd = (int)KPGD;
        	    break;
        	case VK_LEFT:
        	    win_kbd = (int)KLEFT;
        	    break;
        	case VK_RIGHT:
        	    win_kbd = (int)KRIGHT;
        	    break;
        	case VK_UP:
        	    win_kbd = (int)KUP;
        	    break;
        	case VK_DOWN:
        	    win_kbd = (int)KDOWN;
        	    break;
            case VK_HOME:
                win_kbd = (int)KHOME;
                break;
            case VK_BACK:
                win_kbd = (int)KBACK;
                break;
    	   }
    	   return 0;
    	   break;
    }
    return DefWindowProc(hwnd, nMsg, wParam, lParam);
}

void SIGNAL_RESIZE(int n){
}

uiLowLevel::uiLowLevel(){
    memset(this, 0, sizeof(uiLowLevel));
    InitScreen();
}

uiLowLevel::~uiLowLevel(){
}

HWND hwndMain;
WNDCLASSEX wndclass;
char *szMainWndClass = "hxe [Hexadecimal File Editor]";
void uiLowLevel::InitScreen(){
    HICON hico = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(200));
    memset(&wndclass, 0, sizeof(WNDCLASSEX));
    wndclass.lpszClassName = szMainWndClass;
    wndclass.cbSize = sizeof(WNDCLASSEX);
    wndclass.style = CS_HREDRAW | CS_VREDRAW;
    wndclass.lpfnWndProc = MainWndProc;
    wndclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
    wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
    wndclass.hIcon = hico;
    wndclass.hIconSm = hico;
    RegisterClassEx(&wndclass);

    hwndMain =
	CreateWindow(szMainWndClass, szMainWndClass, WS_OVERLAPPEDWINDOW,
		     CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		     CW_USEDEFAULT, NULL, NULL, NULL, NULL);
    if (hwndMain) {
    	ShowWindow(hwndMain, SW_SHOWNORMAL);
    	cursor_on(1);
    	UpdateWindow(hwndMain);
    }
}

void uiLowLevel::commit(){
}

void uiLowLevel::clrscr(){
}

void uiLowLevel::clrln(){
}

long uiLowLevel::GetMaxX(){
    long retval = 0;
    TEXTMETRIC tm;
    RECT rc;
    HGDIOBJ oldobj;
    HDC hdc;
        hdc = GetDC(hwndMain);
        oldobj = SelectObject(hdc, GetStockObject(ANSI_FIXED_FONT));
        GetTextMetrics(hdc, &tm);
        GetClientRect(hwndMain, &rc);
        SelectObject(hdc, oldobj);
        retval = (int) ((rc.right - rc.left) / tm.tmMaxCharWidth);
        ReleaseDC(hwndMain, hdc);
    return retval;
}

long uiLowLevel::GetMaxY(){
    long retval = 0;
    TEXTMETRIC tm;
    RECT rc;
    HGDIOBJ oldobj;
    HDC hdc;
        hdc = GetDC(hwndMain);
        oldobj = SelectObject(hdc, GetStockObject(ANSI_FIXED_FONT));
        GetTextMetrics(hdc, &tm);
        GetClientRect(hwndMain, &rc);
        SelectObject(hdc, oldobj);
        retval = (rc.bottom - rc.top) / (tm.tmHeight);
        ReleaseDC(hwndMain, hdc);
    return retval;
}

void uiLowLevel::cursor_on(int sw_turn){
    TEXTMETRIC tm;
    HGDIOBJ oldobj;
    HDC hdc;
        if(!SW_GOTFOCUS) sw_turn = 0;
        if (sw_turn) {
        	hdc = GetDC(hwndMain);
         	oldobj = SelectObject(hdc, GetStockObject(ANSI_FIXED_FONT));
        	GetTextMetrics(hdc, &tm);
        	SelectObject(hdc, oldobj);
        	//CreateCaret(hwndMain, (HBITMAP)0, tm.tmMaxCharWidth, tm.tmHeight);
        	CreateCaret(hwndMain, (HBITMAP)0, tm.tmMaxCharWidth, CURSOR_HEIGHT * -1);
        	ShowCaret(hwndMain);
        	ReleaseDC(hwndMain, hdc);
        } else {
        	CreateCaret(hwndMain, (HBITMAP)0, 0, 0);
    	    HideCaret(hwndMain);
        }
}


void uiLowLevel::gotoxy(int x, int y){
    int _y, _x;
    TEXTMETRIC tm;
    HGDIOBJ oldobj;
    HDC hdc;
        hdc = GetDC(hwndMain);
        oldobj = SelectObject(hdc, GetStockObject(ANSI_FIXED_FONT));
        GetTextMetrics(hdc, &tm);
        SelectObject(hdc, oldobj);
        _y = tm.tmHeight * y;
        _x = tm.tmMaxCharWidth * x;
        SetCaretPos(_x, _y + tm.tmHeight);
        ReleaseDC(hwndMain, hdc);
}


void uiLowLevel::setDisplayModus(DISPLAY_MODUS modus){
    mModus = modus;
}

void uiLowLevel::printchar(const char c){
    int _y, _x, len = 1;
    POINT p;
    TEXTMETRIC tm;
    HGDIOBJ oldobj;
    HDC hdc;

        HideCaret(hwndMain);
        hdc = GetDC(hwndMain);
        oldobj = SelectObject(hdc, GetStockObject(ANSI_FIXED_FONT));
        GetTextMetrics(hdc, &tm);
        GetCaretPos(&p);
        p.y -= tm.tmHeight;

        SetBkMode(hdc, OPAQUE);
        switch(mModus){
            case DISP_REVERSE:
            	SetTextColor(hdc, COL_BACK);
            	SetBkColor(hdc, COL_FORE);
                break;
            case DISP_HIGHLIGHTED:
            	SetTextColor(hdc, COL_HIGH);
            	SetBkColor(hdc, COL_BACK);
                break;
            case DISP_NORMAL:
            	SetTextColor(hdc, COL_FORE);
            	SetBkColor(hdc, COL_BACK);
                break;
        };
        TextOut(hdc, p.x, p.y, &c, len);
        _y = (long) (p.y / tm.tmHeight);
        _x = (long) (p.x / tm.tmMaxCharWidth) + 1;
        gotoxy(_x, _y);
        SelectObject(hdc, oldobj);
        ShowCaret(hwndMain);
        ReleaseDC(hwndMain, hdc);
}
/*
void uiLowLevel::print(const char *str){
    int _y, _x, len, maxx;
    POINT p;
    TEXTMETRIC tm;
    HGDIOBJ oldobj;
    HDC hdc;
        HideCaret(hwndMain);
        maxx = GetMaxX();
        len = strlen(str);
        hdc = GetDC(hwndMain);
        oldobj = SelectObject(hdc, GetStockObject(ANSI_FIXED_FONT));
        GetTextMetrics(hdc, &tm);
        GetCaretPos(&p);
        p.y -= tm.tmHeight;

        if(len > maxx)len = maxx;
            SetBkMode(hdc, OPAQUE);
            switch(mModus){
                case DISP_REVERSE:
                	SetTextColor(hdc, COL_BACK);
                	SetBkColor(hdc, COL_FORE);
                    break;
                case DISP_HIGHLIGHTED:
                	SetTextColor(hdc, COL_HIGH);
                	SetBkColor(hdc, COL_BACK);
                    break;
                case DISP_NORMAL:
                	SetTextColor(hdc, COL_FORE);
                	SetBkColor(hdc, COL_BACK);
                    break;
            };
            TextOut(hdc, p.x, p.y, str, len);
        _y = (long) (p.y/ tm.tmHeight);
        _x = (long) (p.x / tm.tmMaxCharWidth) + len;
        gotoxy(_x, _y);
        SelectObject(hdc, oldobj);
        ShowCaret(hwndMain);
        ReleaseDC(hwndMain, hdc);
}
*/

void uiLowLevel::print(const char *str){
    int _y, _x, maxx;
    POINT p;
    TEXTMETRIC tm;
    HGDIOBJ oldobj;
    HDC hdc;
    char *from = (char*)str;
    unsigned long len, sze, maxy;
        HideCaret(hwndMain);
        maxx = GetMaxX();
        maxy = GetMaxY();
        len = strlen(str);
        hdc = GetDC(hwndMain);
        oldobj = SelectObject(hdc, GetStockObject(ANSI_FIXED_FONT));
        GetTextMetrics(hdc, &tm);
        GetCaretPos(&p);
        p.y -= tm.tmHeight;
        maxy *= tm.tmHeight;

        SetBkMode(hdc, OPAQUE);
        switch(mModus){
            case DISP_REVERSE:
            	SetTextColor(hdc, COL_BACK);
            	SetBkColor(hdc, COL_FORE);
                break;
            case DISP_HIGHLIGHTED:
            	SetTextColor(hdc, COL_HIGH);
            	SetBkColor(hdc, COL_BACK);
                break;
            case DISP_NORMAL:
            	SetTextColor(hdc, COL_FORE);
            	SetBkColor(hdc, COL_BACK);
                break;
        };
        //if(len > maxx)len = maxx;
        while(len){
            if(len > (unsigned long)maxx) sze = maxx; else sze = len;
            TextOut(hdc, p.x, p.y, from, sze);
            from += sze;
            if(sze == (unsigned long)maxx){
              p.y += tm.tmHeight;
              p.x = 0;
              sze = 0;
            }
            if((unsigned long)p.y < maxy){
                _y = (long) (p.y / tm.tmHeight);
                _x = (long) (p.x / tm.tmMaxCharWidth) + sze;
                gotoxy(_x, _y);
            }else
                from += strlen(from);
            len = strlen(from);
        }

        SelectObject(hdc, oldobj);
        ShowCaret(hwndMain);
        ReleaseDC(hwndMain, hdc);
}

//safe formatted output
void uiLowLevel::printxy(int x, int y, const char *str){
     gotoxy(x, y);
     print(str);
}

void uiLowLevel::printcharxy(int x, int y, const char c){
    if(x > GetMaxX()) x = GetMaxX();
    if(y > GetMaxX()) y = GetMaxY();
    gotoxy(x, y);
    printchar(c);
}

/*
    The non stock object "Terminal" font in windows shows the ASCII + EXTENDED codes
    using this font would make the GetDisplayChar obsolete...

    Converts a char to a '.' if it cannot be displayed on the screen
*/
char uiLowLevel::GetDisplayChar(char ch){
    if( (unsigned int)ch >= ' '  && (unsigned int)ch <= ']'
        ||
        (unsigned int)ch >= '_' && (unsigned int)ch <= '}'
    ) ch = ch; else ch = '.';
    return ch;
}
